import os
import sys
from os.path import dirname
from unittest import TestCase

from typing_extensions import List, Dict, Union, Tuple

from ripple_down_rules.utils import extract_imports, make_set, stringify_hint, get_relative_import, \
    get_imports_from_types


class UtilsTestCase(TestCase):

    def test_extract_imports_from_file(self):
        # Test with a file that has imports
        file_path = "test_file.py"
        with open(file_path, "w") as f:
            f.write("import os\n")
            f.write("from module import function\n")
            f.write("from ripple_down_rules.utils import make_set\n")
            f.write("print('Hello World')\n")

        expected_scope = {"os": os, "make_set": make_set}
        actual_imports = extract_imports(file_path)
        self.assertEqual(expected_scope, actual_imports)

        # Clean up
        os.remove(file_path)


    def test_stringify_hint(self):

        self.assertEqual(stringify_hint(int), "int")
        self.assertEqual(stringify_hint(str), "str")
        self.assertEqual(stringify_hint(List[int]), "List[int]")
        self.assertEqual(stringify_hint(Dict[str, Union[int, str]]), "Dict[str, Union[int, str]]")
        self.assertEqual(stringify_hint(None), "None")
        self.assertEqual(stringify_hint("CustomType"), "CustomType")
        self.assertEqual(stringify_hint(List[Dict[str, Union[int, str]]]), "List[Dict[str, Union[int, str]]]")
        self.assertEqual(stringify_hint(List[Dict[str, Union[int, List[Tuple[str, float]]]]]),
                                        "List[Dict[str, Union[int, List[Tuple[str, float]]]]]")

    def test_get_relative_import(self):
        package_dir = f"{dirname(__file__)}/../src/ripple_down_rules"

        target_file = os.path.join(package_dir, "datastructures", "case.py")
        imported_module = os.path.join(package_dir, "rdr.py")

        rel_import = get_relative_import(target_file, imported_module)
        expected_import = "..rdr"
        assert rel_import == expected_import

        imported_module = os.path.join(package_dir, "datastructures", "case.py")
        target_file = os.path.join(package_dir, "rdr.py")

        rel_import = get_relative_import(target_file, imported_module)
        expected_import = ".datastructures.case"
        assert rel_import == expected_import

        target_file = os.path.join(package_dir, "datastructures", "nested_data", "case.py")
        imported_module = os.path.join(package_dir, "rdr.py")

        rel_import = get_relative_import(target_file, imported_module)
        expected_import = "...rdr"
        assert rel_import == expected_import

    def test_get_imports_from_types(self):
        from .datasets import World, Species

        imports = get_imports_from_types([World, Species])

        assert imports == ["from test.datasets import Species, World"]

        from ripple_down_rules.rdr import GeneralRDR

        package_dir = dirname(sys.modules["ripple_down_rules"].__file__)
        target_file = os.path.join(package_dir, "datastructures", "case.py")
        imports = get_imports_from_types([GeneralRDR], target_file, "ripple_down_rules")
        assert imports == ["from ..rdr import GeneralRDR"]



import unittest

from sqlalchemy import select

from .datasets import *


class TestDDL(unittest.TestCase):

    session: sqlalchemy.orm.Session

    def setUp(self):
        engine = sqlalchemy.create_engine("sqlite:///:memory:")
        Base.metadata.create_all(engine)
        self.session = sqlalchemy.orm.Session(engine)

    def test_animal(self):
        a1 = MappedAnimal(name="animal", hair=True, feathers=False, eggs=True, milk=False, airborne=False, aquatic=False, predator=False,
                          toothed=True, backbone=True, breathes=True, venomous=False, fins=False, legs=4, tail=True,
                          domestic=False, catsize=True, species=Species.mammal)
        # print(a1.habitats)
        # print(isinstance(a1.habitats, set))
        self.session.add(a1)
        self.session.commit()

        animals = self.session.scalars(select(MappedAnimal)).all()

        a1.habitats.add(HabitatTable(Habitat.water))
        # print(a1.habitats)
        self.session.commit()

        queried_a1 = self.session.scalars(select(HabitatTable)).first()
        # print(queried_a1)




"""
This file contains decorators for the RDR (Ripple Down Rules) framework. Where each type of RDR has a decorator
that can be used with any python function such that this function can benefit from the incremental knowledge acquisition
of the RDRs.
"""
import os.path
from functools import wraps

from typing_extensions import Callable, Optional, Type, Tuple, Dict, Any, Self, get_type_hints, List, Union, Sequence

from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.experts import Expert, Human
from ripple_down_rules.rdr import GeneralRDR
try:
    from ripple_down_rules.user_interface.gui import RDRCaseViewer
except ImportError:
    RDRCaseViewer = None
from ripple_down_rules.utils import get_method_args_as_dict, get_func_rdr_model_name, make_set, \
    get_method_class_if_exists, str_to_snake_case


class RDRDecorator:
    rdr: GeneralRDR

    def __init__(self, models_dir: str,
                 output_type: Tuple[Type],
                 mutual_exclusive: bool,
                 output_name: str = "output_",
                 fit: bool = True,
                 expert: Optional[Expert] = None,
                 update_existing_rules: bool = True,
                 viewer: Optional[RDRCaseViewer] = None,
                 package_name: Optional[str] = None,
                 use_generated_classifier: bool = False,
                 ask_now: Callable[[Any, Any], bool] = lambda _,n: True,
                 fitting_decorator: Optional[Callable] = None):
        """
        :param models_dir: The directory to save/load the RDR models.
        :param output_type: The type of the output. This is used to create the RDR model.
        :param mutual_exclusive: If True, the output types are mutually exclusive.
         If None, the RDR model will not be saved as a python file.
        :param output_name: The name of the output. This is used to create the RDR model.
        :param fit: If True, the function will be in fit mode. This means that the RDR will prompt the user for the
            correct output if the function's output is not in the RDR model. If False, the function will be in
            classification mode. This means that the RDR will classify the function's output based on the RDR model.
        :param expert: The expert that will be used to prompt the user for the correct output. If None, a Human
            expert will be used.
        :param update_existing_rules: If True, the function will update the existing RDR rules
         even if they gave an output.
        :param viewer: The viewer to use for the RDR model. If None, no viewer will be used.
        :param package_name: The package name to use for relative imports in the RDR model.
        :param use_generated_classifier: If True, the function will use the generated classifier instead of the RDR model.
        :return: A decorator to use a GeneralRDR as a classifier that monitors and modifies the function's output.
        """
        self.rdr_models_dir = models_dir
        self.model_name: Optional[str] = None
        self.output_type = output_type
        self.parsed_output_type: List[Type] = []
        self.mutual_exclusive = mutual_exclusive
        self.output_name = output_name
        self.fit: bool = fit
        self.expert: Optional[Expert] = expert
        self.update_existing_rules = update_existing_rules
        self.viewer = viewer
        self.package_name = package_name
        self.use_generated_classifier = use_generated_classifier
        self.generated_classifier: Optional[Callable] = None
        self.ask_now = ask_now
        self.fitting_decorator = fitting_decorator
        self.load()

    def decorator(self, func: Callable) -> Callable:

        @wraps(func)
        def wrapper(*args, **kwargs) -> Optional[Any]:

            if self.model_name is None:
                self.initialize_rdr_model_name_and_load(func)

            func_output = {self.output_name: func(*args, **kwargs)}

            case, case_dict = self.create_case_from_method(func, func_output, *args, **kwargs)

            @self.fitting_decorator
            def fit():
                if len(self.parsed_output_type) == 0:
                    self.parsed_output_type = self.parse_output_type(func, self.output_type, *args)
                if self.expert is None:
                    self.expert = Human(viewer=self.viewer,
                                        answers_save_path=self.rdr_models_dir + f'/{self.model_name}/expert_answers')
                case_query = self.create_case_query_from_method(func, func_output,
                                                                self.parsed_output_type,
                                                                self.mutual_exclusive,
                                                                case, case_dict,
                                                                *args, **kwargs)
                output = self.rdr.fit_case(case_query, expert=self.expert,
                                           update_existing_rules=self.update_existing_rules,
                                           viewer=self.viewer)
                return output
            
            if self.fit and not self.use_generated_classifier and self.ask_now(case_dict, self.rdr.classify(case)):
                output = fit()
            else:
                if self.use_generated_classifier:
                    if self.generated_classifier is None:
                        model_path = os.path.join(self.rdr_models_dir, self.model_name)
                        self.generated_classifier = self.rdr.get_rdr_classifier_from_python_file(model_path)
                    output = self.generated_classifier(case)
                else:
                    output = self.rdr.classify(case)

            if self.output_name in output:
                return output[self.output_name]
            else:
                return func_output[self.output_name]

        return wrapper

    @staticmethod
    def create_case_query_from_method(func: Callable,
                                      func_output: Dict[str, Any],
                                      output_type: Sequence[Type],
                                      mutual_exclusive: bool,
                                      case: Optional[Case] = None,
                                      case_dict: Optional[Dict[str, Any]] = None,
                                      *args, **kwargs) -> CaseQuery:
        """
        Create a CaseQuery from the function and its arguments.

        :param func: The function to create a case from.
        :param func_output: The output of the function as a dictionary, where the key is the output name.
        :param output_type: The type of the output as a sequence of types.
        :param mutual_exclusive: If True, the output types are mutually exclusive.
        :param args: The positional arguments of the function.
        :param kwargs: The keyword arguments of the function.
        :return: A CaseQuery object representing the case.
        """
        output_type = make_set(output_type)
        if case is None or case_dict is None:
            case, case_dict = RDRDecorator.create_case_from_method(func, func_output, *args, **kwargs)
        scope = func.__globals__
        scope.update(case_dict)
        func_args_type_hints = get_type_hints(func)
        output_name = list(func_output.keys())[0]
        func_args_type_hints.update({output_name: Union[tuple(output_type)]})
        return CaseQuery(case, output_name, tuple(output_type),
                         mutual_exclusive, scope=scope,
                         is_function=True, function_args_type_hints=func_args_type_hints)

    @staticmethod
    def create_case_from_method(func: Callable,
                                func_output: Dict[str, Any],
                                *args, **kwargs) -> Tuple[Case, Dict[str, Any]]:
        """
        Create a Case from the function and its arguments.

        :param func: The function to create a case from.
        :param func_output: A dictionary containing the output of the function, where the key is the output name.
        :param args: The positional arguments of the function.
        :param kwargs: The keyword arguments of the function.
        :return: A Case object representing the case.
        """
        case_dict = get_method_args_as_dict(func, *args, **kwargs)
        case_dict.update(func_output)
        case_name = get_func_rdr_model_name(func)
        return Case(dict, id(case_dict), case_name, case_dict, **case_dict), case_dict

    def initialize_rdr_model_name_and_load(self, func: Callable) -> None:
        model_file_name = get_func_rdr_model_name(func, include_file_name=True)
        self.model_name = str_to_snake_case(model_file_name)
        self.load()

    @staticmethod
    def parse_output_type(func: Callable, output_type: Any, *args) -> List[Type]:
        parsed_output_type = []
        for ot in make_set(output_type):
            if ot is Self:
                func_class = get_method_class_if_exists(func, *args)
                if func_class is not None:
                    parsed_output_type.append(func_class)
                else:
                    raise ValueError(f"The function {func} is not a method of a class,"
                                     f" and the output type is {Self}.")
            else:
                parsed_output_type.append(ot)
        return parsed_output_type

    def save(self):
        """
        Save the RDR model to the specified directory.
        """
        self.rdr.save(self.rdr_models_dir, self.model_name, package_name=self.package_name)

    def load(self):
        """
        Load the RDR model from the specified directory.
        """
        self.rdr = None
        if self.model_name is not None:
            model_path = os.path.join(self.rdr_models_dir, self.model_name + f"/rdr_metadata/{self.model_name}.json")
            if os.path.exists(os.path.join(self.rdr_models_dir, model_path)):
                self.rdr = GeneralRDR.load(self.rdr_models_dir, self.model_name, package_name=self.package_name)
                self.rdr.set_viewer(self.viewer)
        if self.rdr is None:
            self.rdr = GeneralRDR(save_dir=self.rdr_models_dir, model_name=self.model_name,
                                  viewer=self.viewer)

    def update_from_python(self):
        """
        Update the RDR model from a python file.
        """
        self.rdr.update_from_python(self.rdr_models_dir, package_name=self.package_name)



import os
from unittest import skip

import requests
from IPython.terminal.embed import InteractiveShellEmbed
# from openai import OpenAI
# client = OpenAI(api_key=os.getenv("OPENAI_API_KEY"))
from prompt_toolkit.key_binding import KeyBindings
from traitlets.config import Config
import pprint


class ContextAwareCopilotShell:
    def __init__(self, scope=None):
        self.scope = scope or {}
        self.shell = self._init_shell()

    def _init_shell(self):
        cfg = Config()
        shell = InteractiveShellEmbed(config=cfg, user_ns=self.scope)
        self._add_keybindings(shell)
        return shell

    def _add_keybindings(self, shell):
        kb = KeyBindings()

        @kb.add('c-space')  # Trigger with Ctrl+Space
        def suggest(event):
            buffer = event.app.current_buffer
            current_code = buffer.document.text.strip()
            context = self._get_context()

            print("\n🔍 Generating context-aware suggestion...\n")
            # suggestion = self.get_suggestion(current_code, context)
            suggestion = self.get_suggestion_local(current_code, context)
            print(f"\n💡 Suggestion:\n{suggestion}\n")

        shell.pt_app.key_bindings = kb

    def _get_context(self):
        """Create a summary of local variables for the prompt."""
        ctx = self.shell.user_ns.copy()
        summary = []

        for k, v in ctx.items():
            if k.startswith("__") or callable(v):
                continue
            try:
                typename = type(v).__name__
                valstr = pprint.pformat(v, width=40)
                summary.append(f"{k} ({typename}): {valstr}")
            except Exception:
                continue

        return "\n".join(summary[-10:])  # Limit to last 10 vars

    def get_suggestion(self, code, context_vars):
        prompt = f"""You're a Python assistant helping inside a live debugger shell.
Current variables:
{context_vars}

Based on this context, suggest how to complete or improve the following Python input:
{code}

Respond with a Python snippet that makes sense in context.
"""
        try:
            response = client.chat.completions.create(model="gpt-4",
                                                      messages=[
                                                          {"role": "system",
                                                           "content": "You're a helpful Python code assistant."},
                                                          {"role": "user", "content": prompt}
                                                      ],
                                                      max_tokens=150,
                                                      temperature=0.3)
            return response.choices[0].message.content.strip()
        except Exception as e:
            return f"❌ Error: {e}"

    def get_suggestion_local(self, code, context_vars):
        prompt = f"""You are a helpful Python assistant.
    Current variables:
    {context_vars}

    Suggest a Python code snippet to complete this input:
    {code}
    """
        print(f"Prompt:\n{prompt}\n")
        try:
            response = requests.post("http://localhost:11434/api/generate", json={
                "model": "codellama:7b-instruct",  # or "phi"
                "prompt": prompt,
                "stream": False,
            })
            result = response.json()
            return result.get("response", "").strip()
        except Exception as e:
            return f"❌ Local LLM error: {e}"

    def run(self):
        self.shell.user_ns.update(self.scope)
        self.shell()
        self._sync_back()

    def _sync_back(self):
        # Keep scope up-to-date after running
        for k in self.scope:
            self.scope[k] = self.shell.user_ns.get(k, self.scope[k])


def run_shell():
    user = {"name": "Alice", "age": 30}
    data = [3, 1, 4]
    shell = ContextAwareCopilotShell(locals())
    shell.run()


# run_shell()



from __future__ import annotations

import os
from types import ModuleType

from ripple_down_rules.datastructures.dataclasses import CaseFactoryMetaData

from .datastructures.case import create_case
from .datastructures.dataclasses import CaseQuery
from typing_extensions import Type, Optional, Callable, Any, Dict, TYPE_CHECKING, Union

from .utils import get_func_rdr_model_name, copy_case, make_set, update_case
from .utils import calculate_precision_and_recall

if TYPE_CHECKING:
    from .rdr import RippleDownRules


def general_rdr_classify(classifiers_dict: Dict[str, Union[ModuleType, RippleDownRules]],
                         case: Any, modify_original_case: bool = False,
                         case_query: Optional[CaseQuery] = None) -> Dict[str, Any]:
    """
    Classify a case by going through all classifiers and adding the categories that are classified,
     and then restarting the classification until no more categories can be added.

    :param classifiers_dict: A dictionary mapping conclusion types to the classifiers that produce them.
    :param case: The case to classify.
    :param modify_original_case: Whether to modify the original case or create a copy and modify it.
    :param case_query: The case query to extract metadata from if needed.
    :return: The categories that the case belongs to.
    """
    conclusions = {}
    case = create_case(case)
    case_cp = copy_case(case) if not modify_original_case else case
    while True:
        new_conclusions = {}
        for attribute_name, rdr in classifiers_dict.items():
            pred_atts = rdr.classify(case_cp, case_query=case_query)
            if pred_atts is None and type(None) not in rdr.conclusion_type:
                continue
            if rdr.mutually_exclusive:
                if attribute_name not in conclusions or \
                        (attribute_name in conclusions and conclusions[attribute_name] != pred_atts):
                    conclusions[attribute_name] = pred_atts
                    new_conclusions[attribute_name] = pred_atts
            else:
                pred_atts = make_set(pred_atts)
                if attribute_name in conclusions:
                    pred_atts = {p for p in pred_atts if p not in conclusions[attribute_name]}
                if len(pred_atts) > 0:
                    new_conclusions[attribute_name] = pred_atts
                    if attribute_name not in conclusions:
                        conclusions[attribute_name] = set()
                    conclusions[attribute_name].update(pred_atts)
            if attribute_name in new_conclusions:
                temp_case_query = CaseQuery(case_cp, attribute_name, rdr.conclusion_type, rdr.mutually_exclusive)
                update_case(temp_case_query, new_conclusions)
        if len(new_conclusions) == 0: #or (len(classifiers_dict) == 1 and all(rdr.mutually_exclusive for rdr in classifiers_dict.values())):
            break
    return conclusions


def is_matching(classifier: Callable[[Any], Any], case_query: CaseQuery, pred_cat: Optional[Dict[str, Any]] = None) -> bool:
    """
    :param classifier: The RDR classifier to check the prediction of.
    :param case_query: The case query to check.
    :param pred_cat: The predicted category.
    :return: Whether the classifier prediction is matching case_query target or not.
    """
    if case_query.target is None:
        return False
    if pred_cat is None:
        pred_cat = classifier(case_query.case)
    if not isinstance(pred_cat, dict):
        pred_cat = {case_query.attribute_name: pred_cat}
    target = {case_query.attribute_name: case_query.target_value}
    precision, recall = calculate_precision_and_recall(pred_cat, target)
    return all(recall) and all(precision)


def load_or_create_func_rdr_model(func, model_dir: str, rdr_type: Type[RippleDownRules],
                                  **rdr_kwargs) -> RippleDownRules:
    """
    Load the RDR model of the function if it exists, otherwise create a new one.

    :param func: The function to load the model for.
    :param model_dir: The directory where the model is stored.
    :param rdr_type: The type of the RDR model to load.
    :param rdr_kwargs: Additional arguments to pass to the RDR constructor in the case of a new model.
    """
    model_name = get_func_rdr_model_name(func)
    model_path = os.path.join(model_dir, model_name, "rdr_metadata", f"{model_name}.json")
    if os.path.exists(model_path):
        rdr = rdr_type.load(load_dir=model_dir, model_name=model_name)
    else:
        rdr = rdr_type(**rdr_kwargs)
    return rdr



from os.path import dirname, join

from typing_extensions import Union, Any, Dict

from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.rdr_decorators import RDRDecorator
from ripple_down_rules.utils import is_iterable, make_list

# app = QApplication(sys.argv)
save_dir = join(dirname(__file__), '..', 'src', 'ripple_down_rules')
# viewer = RDRCaseViewer(save_dir=save_dir)
viewer = None
rdr_decorator: RDRDecorator = RDRDecorator(save_dir, (bool,), True,
                                           fit=False,
                                           viewer=viewer)


@rdr_decorator.decorator
def should_i_ask_the_expert_for_a_target(conclusions: Union[Any, Dict[str, Any]],
                                         case_query: CaseQuery,
                                         update_existing: bool) -> bool:
    """
    Determine if the rdr should ask the expert for the target of a given case query.

    :param conclusions: The conclusions of the case.
    :param case_query: The query containing the case to classify.
    :param ask_always: Whether to ask the expert always.
    :param update_existing: Whether to update rules that gave the required type of conclusions.
    :return: True if the rdr should ask the expert, False otherwise.
    """
    if conclusions is None:
        return True
    elif is_iterable(conclusions) and len(conclusions) == 0:
        return True
    elif isinstance(conclusions, dict):
        if case_query.attribute_name not in conclusions:
            return True
        conclusions = conclusions[case_query.attribute_name]
    conclusion_types = map(type, make_list(conclusions))
    if not any(ct in case_query.core_attribute_type for ct in conclusion_types):
        return True
    elif update_existing:
        return True
    else:
        return False


def test_should_i_ask_the_expert_for_a_target(conclusions: Union[Any, Dict[str, Any]],
                                              case_query: CaseQuery,
                                              update_existing: bool):
    out = should_i_ask_the_expert_for_a_target(conclusions, case_query, update_existing)
    # if not out:
    #     rdr_decorator.fit = True
    #     should_i_ask_the_expert_for_a_target(conclusions, case_query, ask_always, update_existing)
    # rdr_decorator.fit = False



from ...datasets import World, Handle, Container, FixedConnection, PrismaticConnection
from ...conf.world.base_config import WorldConf, HandleConf, ContainerConf, FixedConnectionConf, PrismaticConnectionConf


def create_world_with_handles_and_containers(world_conf: WorldConf) -> World:
    world = World()

    for body in world_conf.bodies:
        if isinstance(body, HandleConf):
            world.bodies.append(Handle(body.name, world=world))
        elif isinstance(body, ContainerConf):
            world.bodies.append(Container(body.name, world=world))
    for connection in world_conf.connections:
        parent = next((b for b in world.bodies if b.name == connection.parent.name), None)
        child = next((b for b in world.bodies if b.name == connection.child.name), None)
        if parent and child:
            if isinstance(connection, FixedConnectionConf):
                world.connections.append(FixedConnection(parent=parent, child=child, world=world))
            elif isinstance(connection, PrismaticConnectionConf):
                world.connections.append(PrismaticConnection(parent=parent, child=child, world=world))

    return world



import logging

from ripple_down_rules.datastructures.case import Case
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.utils import SubclassJSONSerializer

try:
    import graphviz
except ImportError:
    graphviz = None
    logging.debug("Graphviz is not installed")


def is_simple(obj):
    return isinstance(obj, (int, float, str, bool, type(None)))


def get_colored_value(value):
    if isinstance(value, str):
        color = '#A31515'  # red for strings
        val = repr(value)
    elif isinstance(value, (int, float)):
        color = '#098658'  # green for numbers
        val = str(value)
    elif isinstance(value, bool):
        color = '#0000FF'  # blue for booleans
        val = str(value)
    elif value is None:
        color = '#808080'  # gray for None
        val = 'None'
    else:
        color = '#000000'  # fallback
        val = str(value)
    val = (val.replace("class ", "").replace("'", "").replace('<', "")
           .replace('>', "").replace('?', '').replace('\n', ' '))
    val = val[:50] + '...' if len(val) > 50 else val
    return f'<FONT COLOR="{color}">{val}</FONT>'


def generate_object_graph(obj, name='root', seen=None, graph=None, current_depth=0, max_depth=3, chain_name=None,
                          included_attrs=None, iterable_limit=3):
    if seen is None:
        seen = set()
    if graph is None:
        graph = graphviz.Digraph(format='svg')
        graph.attr('node', shape='plaintext')

    obj_id = id(obj)
    if obj_id in seen or current_depth > max_depth:
        return graph
    seen.add(obj_id)

    # Build HTML table label for this object node
    type_name = type(obj).__name__ if not isinstance(obj, type) else obj.__name__
    rows = [f'<TR><TD><B>{name}</B></TD><TD><I>{type_name}</I></TD></TR>']

    # We'll keep track of non-simple attrs to add edges later
    non_simple_attrs = []

    if isinstance(obj, (list, tuple, set, dict)):
        items = obj.items() if isinstance(obj, dict) else enumerate(obj)
        i = 0
        for idx, item in items:
            if idx == "scope":
                continue
            if i >= iterable_limit:
                rows.append(f'<TR><TD ALIGN="LEFT" PORT="{idx}">...</TD><TD ALIGN="LEFT">...</TD></TR>')
                break
            # Represent items as attr = index + type (for the label)
            if is_simple(item):
                val_colored = get_colored_value(item)
                rows.append(f'<TR><TD ALIGN="LEFT" PORT="{idx}">[{idx}]</TD><TD ALIGN="LEFT">{val_colored}</TD></TR>')
            else:
                type_name = type(item).__name__ if not isinstance(item, type) else item.__name__
                rows.append(
                    f'<TR><TD ALIGN="LEFT" PORT="{idx}">[{idx}]</TD><TD ALIGN="LEFT"><I>{type_name}</I></TD></TR>')
                non_simple_attrs.append((str(idx), item))
            i += 1

    else:
        for attr in dir(obj):
            if attr.startswith('_'):
                continue
            if isinstance(obj, CaseQuery) and attr == 'scope':
                continue
            if isinstance(obj, Case) and attr in ['data']:
                continue
            if isinstance(obj, SubclassJSONSerializer) and attr == 'data_class_refs':
                continue
            value = getattr(obj, attr)
            if callable(value):
                continue
            if is_simple(value):
                val_colored = get_colored_value(value)
                rows.append(f'<TR><TD ALIGN="LEFT" PORT="{attr}">{attr}</TD><TD ALIGN="LEFT">{val_colored}</TD></TR>')
            else:
                type_name = type(value).__name__ if not isinstance(value, type) else value.__name__
                # Show just name and type inside the node
                rows.append(
                    f'<TR><TD ALIGN="LEFT" PORT="{attr}">{attr}</TD><TD ALIGN="LEFT"><I>{type_name}</I></TD></TR>')
                non_simple_attrs.append((attr, value))

    label = f"""<
        <TABLE BORDER="1" CELLBORDER="1" CELLSPACING="0" CELLPADDING="4">
        {''.join(rows)}
        </TABLE>
    >"""

    graph.node(str(obj_id), label)
    chain_name = chain_name if chain_name is not None else name

    # Add edges from this node to non-simple attribute nodes
    for attr, value in non_simple_attrs:
        if attr.startswith('_'):
            continue
        if attr == 'scope':
            continue
        if callable(value):
            continue
        attr_chain_name = f"{chain_name}.{attr}"
        if included_attrs is not None and attr_chain_name not in included_attrs:
            continue
        generate_object_graph(value, attr, seen, graph, current_depth + 1, max_depth, chain_name=f"{chain_name}.{attr}",
                              included_attrs=included_attrs)
        # Edge from this object's attribute port to nested object's node
        graph.edge(f"{obj_id}:{attr}", str(id(value)), label=attr)

    return graph



from __future__ import annotations

import importlib
import os
from abc import ABC, abstractmethod
from copy import copy
from types import NoneType

from ripple_down_rules.datastructures.dataclasses import CaseFactoryMetaData
from . import logger

try:
    from matplotlib import pyplot as plt

    Figure = plt.Figure
except ImportError as e:
    logger.debug(f"{e}: matplotlib is not installed")
    matplotlib = None
    Figure = None
    plt = None

from sqlalchemy.orm import DeclarativeBase as SQLTable
from typing_extensions import List, Optional, Dict, Type, Union, Any, Self, Tuple, Callable, Set

from .datastructures.callable_expression import CallableExpression
from .datastructures.case import Case, CaseAttribute, create_case
from .datastructures.dataclasses import CaseQuery
from .datastructures.enums import MCRDRMode
from .experts import Expert, Human
from .helpers import is_matching, general_rdr_classify
from .rules import Rule, SingleClassRule, MultiClassTopRule, MultiClassStopRule

try:
    from .user_interface.gui import RDRCaseViewer
except ImportError as e:
    RDRCaseViewer = None
from .utils import draw_tree, make_set, SubclassJSONSerializer, make_list, get_type_from_string, \
    is_conflicting, extract_function_source, extract_imports, get_full_class_name, \
    is_iterable, str_to_snake_case, get_import_path_from_path, get_imports_from_types


class RippleDownRules(SubclassJSONSerializer, ABC):
    """
    The abstract base class for the ripple down rules classifiers.
    """
    fig: Optional[Figure] = None
    """
    The figure to draw the tree on.
    """
    expert_accepted_conclusions: Optional[List[CaseAttribute]] = None
    """
    The conclusions that the expert has accepted, such that they are not asked again.
    """
    _generated_python_file_name: Optional[str] = None
    """
    The name of the generated python file.
    """
    name: Optional[str] = None
    """
    The name of the classifier.
    """
    case_type: Optional[Type] = None
    """
    The type of the case (input) to the RDR classifier.
    """
    case_name: Optional[str] = None
    """
    The name of the case type.
    """
    metadata_folder: str = "rdr_metadata"
    """
    The folder to save the metadata of the RDR classifier.
    """
    model_name: Optional[str] = None
    """
    The name of the model. If None, the model name will be the generated python file name.
    """
    mutually_exclusive: Optional[bool] = None
    """
    Whether the output of the classification of this rdr allows only one possible conclusion or not.
    """

    def __init__(self, start_rule: Optional[Rule] = None, viewer: Optional[RDRCaseViewer] = None,
                 save_dir: Optional[str] = None, model_name: Optional[str] = None):
        """
        :param start_rule: The starting rule for the classifier.
        :param viewer: The viewer gui to use for the classifier. If None, no viewer is used.
        :param save_dir: The directory to save the classifier to.
        """
        self.model_name: Optional[str] = model_name
        self.save_dir = save_dir
        self.start_rule = start_rule
        self.fig: Optional[Figure] = None
        self.viewer: Optional[RDRCaseViewer] = viewer
        if self.viewer is not None:
            self.viewer.set_save_function(self.save)

    def save(self, save_dir: Optional[str] = None, model_name: Optional[str] = None,
             package_name: Optional[str] = None) -> str:
        """
        Save the classifier to a file.

        :param save_dir: The directory to save the classifier to.
        :param model_name: The name of the model to save. If None, a default name is generated.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :return: The name of the saved model.
        """
        save_dir = save_dir or self.save_dir
        if save_dir is None:
            raise ValueError("The save directory cannot be None. Please provide a valid directory to save"
                             " the classifier.")
        if not os.path.exists(save_dir + '/__init__.py'):
            os.makedirs(save_dir, exist_ok=True)
            with open(save_dir + '/__init__.py', 'w') as f:
                f.write("from . import *\n")
        if model_name is not None:
            self.model_name = model_name
        elif self.model_name is None:
            self.model_name = self.generated_python_file_name
        model_dir = os.path.join(save_dir, self.model_name)
        os.makedirs(model_dir, exist_ok=True)
        json_dir = os.path.join(model_dir, self.metadata_folder)
        os.makedirs(json_dir, exist_ok=True)
        self.to_json_file(os.path.join(json_dir, self.model_name))
        self._write_to_python(model_dir, package_name=package_name)
        return self.model_name

    @classmethod
    def load(cls, load_dir: str, model_name: str,
             package_name: Optional[str] = None) -> Self:
        """
        Load the classifier from a file.

        :param load_dir: The path to the model directory to load the classifier from.
        :param model_name: The name of the model to load.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        model_dir = os.path.join(load_dir, model_name)
        json_file = os.path.join(model_dir, cls.metadata_folder, model_name)
        rdr = cls.from_json_file(json_file)
        rdr.save_dir = load_dir
        rdr.model_name = model_name
        try:
            rdr.update_from_python(model_dir, package_name=package_name)
            rdr.to_json_file(json_file)
        except (FileNotFoundError, ValueError, SyntaxError) as e:
            logger.warning(f"Could not load the python file for the model {model_name} from {model_dir}. "
                           f"Make sure the file exists and is valid.")
            rdr.save(save_dir=load_dir, model_name=model_name, package_name=package_name)
        return rdr

    @abstractmethod
    def _write_to_python(self, model_dir: str, package_name: Optional[str] = None):
        """
        Write the tree of rules as source code to a file.

        :param model_dir: The path to the directory to write the source code to.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        pass

    def set_viewer(self, viewer: RDRCaseViewer):
        """
        Set the viewer for the classifier.

        :param viewer: The viewer to set.
        """
        self.viewer = viewer
        if self.viewer is not None:
            self.viewer.set_save_function(self.save)

    def fit(self, case_queries: List[CaseQuery],
            expert: Optional[Expert] = None,
            n_iter: int = None,
            animate_tree: bool = False,
            **kwargs_for_fit_case):
        """
        Fit the classifier to a batch of cases and categories.

        :param case_queries: The cases and categories to fit the classifier to.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param n_iter: The number of iterations to fit the classifier for.
        :param animate_tree: Whether to draw the tree while fitting the classifier.
        :param kwargs_for_fit_case: The keyword arguments to pass to the fit_case method.
        """
        targets = []
        if animate_tree:
            if plt is None:
                raise ImportError("matplotlib is not installed, cannot animate the tree.")
            plt.ion()
        i = 0
        stop_iterating = False
        num_rules: int = 0
        while not stop_iterating:
            for case_query in case_queries:
                pred_cat = self.fit_case(case_query, expert=expert, **kwargs_for_fit_case)
                if case_query.target is None:
                    continue
                target = {case_query.attribute_name: case_query.target(case_query.case)}
                if len(targets) < len(case_queries):
                    targets.append(target)
                match = is_matching(self.classify, case_query, pred_cat)
                if not match:
                    print(f"Predicted: {pred_cat} but expected: {target}")
                if animate_tree and len(self.start_rule.descendants) > num_rules:
                    num_rules = len(self.start_rule.descendants)
                    self.update_figures()
            i += 1
            all_predictions = [1 if is_matching(self.classify, case_query) else 0 for case_query in case_queries
                               if case_query.target is not None]
            all_pred = sum(all_predictions)
            logger.info(f"Accuracy: {all_pred}/{len(targets)}")
            all_predicted = targets and all_pred == len(targets)
            num_iter_reached = n_iter and i >= n_iter
            stop_iterating = all_predicted or num_iter_reached
            if stop_iterating:
                break
        logger.info(f"Finished training in {i} iterations")
        if animate_tree:
            plt.ioff()
            plt.show()

    def __call__(self, case: Union[Case, SQLTable]) -> Union[CallableExpression, Dict[str, CallableExpression]]:
        return self.classify(case)

    @abstractmethod
    def classify(self, case: Union[Case, SQLTable], modify_case: bool = False,
                 case_query: Optional[CaseQuery] = None) \
            -> Optional[Union[CallableExpression, Dict[str, CallableExpression]]]:
        """
        Classify a case.

        :param case: The case to classify.
        :param modify_case: Whether to modify the original case attributes with the conclusion or not.
        :param case_query: The case query containing the case to classify and the target category to compare the case with.
        :return: The category that the case belongs to.
        """
        pass

    def fit_case(self, case_query: CaseQuery,
                 expert: Optional[Expert] = None,
                 update_existing_rules: bool = True,
                 scenario: Optional[Callable] = None,
                 ask_now: Callable = lambda _: True,
                 **kwargs) \
            -> Union[CallableExpression, Dict[str, CallableExpression]]:
        """
        Fit the classifier to a case and ask the expert for refinements or alternatives if the classification is
        incorrect by comparing the case with the target category.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param update_existing_rules: Whether to update the existing same conclusion type rules that already gave
        some conclusions with the type required by the case query.
        :param scenario: The scenario at which the case was created, this is used to recreate the case if needed.
        :ask_now: Whether to ask the expert for refinements or alternatives.
        :return: The category that the case belongs to.
        """
        if case_query is None:
            raise ValueError("The case query cannot be None.")

        self.name = case_query.attribute_name if self.name is None else self.name
        self.case_type = case_query.case_type if self.case_type is None else self.case_type
        self.case_name = case_query.case_name if self.case_name is None else self.case_name
        case_query.scenario = scenario if case_query.scenario is None else case_query.scenario

        expert = expert or Human(viewer=self.viewer,
                                 answers_save_path=self.save_dir + '/expert_answers'
                                 if self.save_dir else None)
        if case_query.target is None:
            case_query_cp = copy(case_query)
            conclusions = self.classify(case_query_cp.case, modify_case=True, case_query=case_query_cp)
            if (self.should_i_ask_the_expert_for_a_target(conclusions, case_query_cp, update_existing_rules)
                    and ask_now(case_query_cp.case)):
                expert.ask_for_conclusion(case_query_cp)
                case_query.target = case_query_cp.target
            if case_query.target is None:
                return self.classify(case_query.case)

        self.update_start_rule(case_query, expert)

        fit_case_result = self._fit_case(case_query, expert=expert, **kwargs)

        if self.save_dir is not None:
            self.save()
            expert.clear_answers()

        return fit_case_result

    @staticmethod
    def should_i_ask_the_expert_for_a_target(conclusions: Union[Any, Dict[str, Any]],
                                             case_query: CaseQuery,
                                             update_existing: bool) -> bool:
        """
        Determine if the rdr should ask the expert for the target of a given case query.

        :param conclusions: The conclusions of the case.
        :param case_query: The query containing the case to classify.
        :param update_existing: Whether to update rules that gave the required type of conclusions.
        :return: True if the rdr should ask the expert, False otherwise.
        """
        if conclusions is None and type(None) not in case_query.core_attribute_type:
            return True
        elif is_iterable(conclusions) and len(conclusions) == 0:
            return True
        elif isinstance(conclusions, dict):
            if case_query.attribute_name not in conclusions:
                return True
            conclusions = conclusions[case_query.attribute_name]
        conclusion_types = map(type, make_list(conclusions))
        if not any(ct in case_query.core_attribute_type for ct in conclusion_types):
            return True
        elif update_existing:
            return True
        else:
            return False

    @abstractmethod
    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CallableExpression, Dict[str, CallableExpression]]:
        """
        Fit the RDR on a case, and ask the expert for refinements or alternatives if the classification is incorrect by
        comparing the case with the target category.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        pass

    @abstractmethod
    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        pass

    def update_figures(self):
        """
        Update the figures of the classifier.
        """
        if isinstance(self, GeneralRDR):
            for i, (rdr_name, rdr) in enumerate(self.start_rules_dict.items()):
                if not rdr.fig:
                    rdr.fig = plt.figure(f"Rule {i}: {rdr_name}")
                draw_tree(rdr.start_rule, rdr.fig)
        else:
            if not self.fig:
                self.fig = plt.figure(0)
            draw_tree(self.start_rule, self.fig)

    @property
    def type_(self):
        return self.__class__

    @property
    def generated_python_file_name(self) -> str:
        if self._generated_python_file_name is None:
            self._generated_python_file_name = self._default_generated_python_file_name
        return self._generated_python_file_name

    @generated_python_file_name.setter
    def generated_python_file_name(self, value: str):
        """
        Set the generated python file name.
        :param value: The new value for the generated python file name.
        """
        self._generated_python_file_name = value

    @property
    @abstractmethod
    def _default_generated_python_file_name(self) -> str:
        """
        :return: The default generated python file name.
        """
        pass

    @abstractmethod
    def update_from_python(self, model_dir: str, package_name: Optional[str] = None):
        """
        Update the rules from the generated python file, that might have been modified by the user.

        :param model_dir: The directory where the generated python file is located.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        pass

    @classmethod
    def get_acronym(cls) -> str:
        """
        :return: The acronym of the classifier.
        """
        if cls.__name__ == "GeneralRDR":
            return "RDR"
        elif cls.__name__ == "MultiClassRDR":
            return "MCRDR"
        else:
            return "SCRDR"

    def get_rdr_classifier_from_python_file(self, package_name: str) -> Callable[[Any], Any]:
        """
        :param package_name: The name of the package that contains the RDR classifier function.
        :return: The module that contains the rdr classifier function.
        """
        # remove from imports if exists first
        package_name = get_import_path_from_path(package_name)
        name = f"{package_name}.{self.generated_python_file_name}" if package_name else self.generated_python_file_name
        module = importlib.import_module(name)
        importlib.reload(module)
        return module.classify


class RDRWithCodeWriter(RippleDownRules, ABC):

    def update_from_python(self, model_dir: str, package_name: Optional[str] = None):
        """
        Update the rules from the generated python file, that might have been modified by the user.

        :param model_dir: The directory where the generated python file is located.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        rules_dict = {r.uid: r for r in [self.start_rule] + list(self.start_rule.descendants)
                      if r.conditions is not None}
        condition_func_names = [f'conditions_{rid}' for rid in rules_dict.keys()]
        conclusion_func_names = [f'conclusion_{rid}' for rid in rules_dict.keys()
                                 if not isinstance(rules_dict[rid], MultiClassStopRule)]
        all_func_names = condition_func_names + conclusion_func_names
        filepath = f"{model_dir}/{self.generated_python_defs_file_name}.py"
        cases_path = f"{model_dir}/{self.generated_python_cases_file_name}.py"
        cases_import_path = get_import_path_from_path(model_dir)
        cases_import_path = f"{cases_import_path}.{self.generated_python_cases_file_name}" if cases_import_path \
            else self.generated_python_cases_file_name
        functions_source = extract_function_source(filepath, all_func_names, include_signature=False)
        # get the scope from the imports in the file
        scope = extract_imports(filepath, package_name=package_name)
        for rule in [self.start_rule] + list(self.start_rule.descendants):
            if rule.conditions is not None:
                rule.conditions.user_input = functions_source[f"conditions_{rule.uid}"]
                rule.conditions.scope = scope
                if os.path.exists(cases_path):
                    module = importlib.import_module(cases_import_path, package=package_name)
                    importlib.reload(module)
                    rule.corner_case_metadata = module.__dict__.get(f"corner_case_{rule.uid}", None)
            if rule.conclusion is not None and not isinstance(rule, MultiClassStopRule):
                rule.conclusion.user_input = functions_source[f"conclusion_{rule.uid}"]
                rule.conclusion.scope = scope

    @abstractmethod
    def write_rules_as_source_code_to_file(self, rule: Rule, file, parent_indent: str = "",
                                           defs_file: Optional[str] = None, cases_file: Optional[str] = None,
                                           package_name: Optional[str] = None):
        """
        Write the rules as source code to a file.

        :param rule: The rule to write as source code.
        :param file: The file to write the source code to.
        :param parent_indent: The indentation of the parent rule.
        :param defs_file: The file to write the definitions to.
        :param cases_file: The file to write the cases to.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        pass

    def _write_to_python(self, model_dir: str, package_name: Optional[str] = None):
        """
        Write the tree of rules as source code to a file.

        :param model_dir: The path to the directory to write the source code to.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        # Make sure the model directory exists and create an __init__.py file if it doesn't exist
        os.makedirs(model_dir, exist_ok=True)
        if not os.path.exists(model_dir + '/__init__.py'):
            with open(model_dir + '/__init__.py', 'w') as f:
                f.write("from . import *\n")

        # Set the file names for the generated python files
        file_name = model_dir + f"/{self.generated_python_file_name}.py"
        defs_file_name = model_dir + f"/{self.generated_python_defs_file_name}.py"
        cases_file_name = model_dir + f"/{self.generated_python_cases_file_name}.py"

        # Get the required imports for the main file and the defs file
        main_types, defs_types, corner_cases_types = self._get_types_to_import()
        imports = get_imports_from_types(main_types, file_name, package_name)
        defs_imports = get_imports_from_types(defs_types, defs_file_name, package_name)
        corner_cases_imports = get_imports_from_types(corner_cases_types, cases_file_name, package_name)

        # Add the imports to the defs file
        with open(defs_file_name, "w") as f:
            f.write('\n'.join(defs_imports) + "\n\n\n")

        # Add the imports to the cases file
        case_factory_import = get_imports_from_types([CaseFactoryMetaData], cases_file_name, package_name)
        corner_cases_imports.extend(case_factory_import)
        with open(cases_file_name, "w") as cases_f:
            cases_f.write("# This file contains the corner cases for the rules.\n")
            cases_f.write('\n'.join(corner_cases_imports) + "\n\n\n")

        # Add the imports, the attributes, and the function definition to the main file
        func_def = f"def classify(case: {self.case_type.__name__}, **kwargs) -> {self.conclusion_type_hint}:\n"
        with open(file_name, "w") as f:
            imports.append(f"from .{self.generated_python_defs_file_name} import *")
            f.write('\n'.join(imports) + "\n\n\n")
            f.write(f"attribute_name = '{self.attribute_name}'\n")
            f.write(f"conclusion_type = ({', '.join([ct.__name__ for ct in self.conclusion_type])},)\n")
            f.write(f"mutually_exclusive = {self.mutually_exclusive}\n")
            f.write(f"\n\n{func_def}")
            f.write(f"{' ' * 4}if not isinstance(case, Case):\n"
                    f"{' ' * 4}    case = create_case(case, max_recursion_idx=3)\n""")

        # Write the rules as source code to the main file
        self.write_rules_as_source_code_to_file(self.start_rule, file_name, " " * 4, defs_file=defs_file_name,
                                                cases_file=cases_file_name, package_name=package_name)

    @property
    @abstractmethod
    def conclusion_type_hint(self) -> str:
        """
        :return: The type hint of the conclusion of the rdr as a string.
        """
        pass

    def _get_types_to_import(self) -> Tuple[Set[Type], Set[Type], Set[Type]]:
        """
        :return: The types of the main, defs, and corner cases files of the RDR classifier that will be imported.
        """
        defs_types = set()
        cases_types = set()
        for rule in [self.start_rule] + list(self.start_rule.descendants):
            if not rule.conditions:
                continue
            for scope in [rule.conditions.scope, rule.conclusion.scope]:
                if scope is None:
                    continue
                defs_types.update(make_set(scope.values()))
            corner_case_types = rule.get_corner_case_types_to_import()
            if corner_case_types is not None:
                cases_types.update(corner_case_types)
        defs_types.add(self.case_type)
        main_types = set()
        main_types.add(self.case_type)
        main_types.update(make_set(self.conclusion_type))
        main_types.update({Union, Optional})
        defs_types.add(Union)
        main_types.update({Case, create_case})
        main_types = main_types.difference(defs_types)
        return main_types, defs_types, cases_types

    @property
    def _default_generated_python_file_name(self) -> Optional[str]:
        """
        :return: The default generated python file name.
        """
        if self.start_rule is None or self.start_rule.conclusion is None:
            return None
        return f"{str_to_snake_case(self.case_name)}_{self.attribute_name}_{self.get_acronym().lower()}"

    @property
    def generated_python_defs_file_name(self) -> str:
        return f"{self.generated_python_file_name}_defs"

    @property
    def generated_python_cases_file_name(self) -> str:
        return f"{self.generated_python_file_name}_cases"

    @property
    def conclusion_type(self) -> Tuple[Type]:
        """
        :return: The type of the conclusion of the RDR classifier.
        """
        all_types = []
        if self.start_rule is not None:
            for rule in [self.start_rule] + list(self.start_rule.descendants):
                all_types.extend(list(rule.conclusion.conclusion_type))
        return tuple(set(all_types))

    @property
    def attribute_name(self) -> str:
        """
        :return: The name of the attribute that the classifier is classifying.
        """
        return self.start_rule.conclusion_name

    def _to_json(self) -> Dict[str, Any]:
        return {"start_rule": self.start_rule.to_json(),
                "generated_python_file_name": self.generated_python_file_name,
                "name": self.name,
                "case_type": get_full_class_name(self.case_type) if self.case_type is not None else None,
                "case_name": self.case_name}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> Self:
        """
        Create an instance of the class from a json
        """
        start_rule = cls.start_rule_type().from_json(data["start_rule"])
        new_rdr = cls(start_rule=start_rule)
        if "generated_python_file_name" in data:
            new_rdr.generated_python_file_name = data["generated_python_file_name"]
        if "name" in data:
            new_rdr.name = data["name"]
        if "case_type" in data:
            new_rdr.case_type = get_type_from_string(data["case_type"])
        if "case_name" in data:
            new_rdr.case_name = data["case_name"]
        return new_rdr

    @staticmethod
    @abstractmethod
    def start_rule_type() -> Type[Rule]:
        """
        :return: The type of the starting rule of the RDR classifier.
        """
        pass


class SingleClassRDR(RDRWithCodeWriter):
    mutually_exclusive: bool = True
    """
    The output of the classification of this rdr negates all other possible outputs, there can only be one true value.
    """

    def __init__(self, default_conclusion: Optional[Any] = None, **kwargs):
        """
        :param start_rule: The starting rule for the classifier.
        :param default_conclusion: The default conclusion for the classifier if no rules fire.
        """
        super(SingleClassRDR, self).__init__(**kwargs)
        self.default_conclusion: Optional[Any] = default_conclusion

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Union[CaseAttribute, CallableExpression, None]:
        """
        Classify a case, and ask the user for refinements or alternatives if the classification is incorrect by
        comparing the case with the target category if provided.

        :param case_query: The case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The category that the case belongs to.
        """
        if case_query.default_value is not None and self.default_conclusion != case_query.default_value:
            self.default_conclusion = case_query.default_value

        pred = self.evaluate(case_query.case)
        if pred.conclusion(case_query.case) != case_query.target_value:
            expert.ask_for_conditions(case_query, pred)
            pred.fit_rule(case_query)

        return self.classify(case_query.case)

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if not self.start_rule:
            expert.ask_for_conditions(case_query)
            self.start_rule = SingleClassRule.from_case_query(case_query)

    def classify(self, case: Case, modify_case: bool = False,
                 case_query: Optional[CaseQuery] = None) -> Optional[Any]:
        """
        Classify a case by recursively evaluating the rules until a rule fires or the last rule is reached.

        :param case: The case to classify.
        :param modify_case: Whether to modify the original case attributes with the conclusion or not.
        :param case_query: The case query containing the case and the target category to compare the case with.
        """
        pred = self.evaluate(case)
        conclusion = pred.conclusion(case) if pred is not None and pred.fired else self.default_conclusion
        if pred is not None and pred.fired and case_query is not None:
            if pred.corner_case_metadata is None and conclusion is not None \
                    and type(conclusion) in case_query.core_attribute_type:
                pred.corner_case_metadata = CaseFactoryMetaData.from_case_query(case_query)
        return conclusion

    def evaluate(self, case: Case) -> SingleClassRule:
        """
        Evaluate the starting rule on a case.
        """
        matched_rule = self.start_rule(case) if self.start_rule is not None else None
        return matched_rule if matched_rule is not None else self.start_rule

    def _write_to_python(self, model_dir: str, package_name: Optional[str] = None):
        super()._write_to_python(model_dir, package_name=package_name)
        with open(model_dir + f"/{self.generated_python_file_name}.py", "a") as f:
            f.write(f"{' ' * 4}else:\n{' ' * 8}return {self.default_conclusion}\n")

    def write_rules_as_source_code_to_file(self, rule: SingleClassRule, filename: str, parent_indent: str = "",
                                           defs_file: Optional[str] = None, cases_file: Optional[str] = None,
                                           package_name: Optional[str] = None):
        """
        Write the rules as source code to a file.
        """
        if rule.conditions:
            rule.write_corner_case_as_source_code(cases_file, package_name=package_name)
            if_clause = rule.write_condition_as_source_code(parent_indent, defs_file)
            with open(filename, "a") as file:
                file.write(if_clause)
            if rule.refinement:
                self.write_rules_as_source_code_to_file(rule.refinement, filename, parent_indent + "    ",
                                                        defs_file=defs_file, cases_file=cases_file,
                                                        package_name=package_name)

            conclusion_call = rule.write_conclusion_as_source_code(parent_indent, defs_file)
            with open(filename, "a") as file:
                file.write(conclusion_call)

            if rule.alternative:
                self.write_rules_as_source_code_to_file(rule.alternative, filename, parent_indent, defs_file=defs_file,
                                                        cases_file=cases_file, package_name=package_name)

    @property
    def conclusion_type_hint(self) -> str:
        all_types = set(list(self.conclusion_type) + [type(self.default_conclusion)])
        if NoneType in all_types:
            return f"Optional[{', '.join([t.__name__ for t in all_types if t is not NoneType])}]"
        return f"Union[{', '.join([t.__name__ for t in all_types])}]"

    def _get_types_to_import(self) -> Tuple[Set[Type], Set[Type], Set[Type]]:
        main_types, def_types, case_types = super()._get_types_to_import()
        main_types.add(type(self.default_conclusion))
        def_types.add(type(self.default_conclusion))
        if self.default_conclusion is None:
            main_types.add(Optional)
            def_types.add(Optional)
        return main_types, def_types, case_types

    @property
    def conclusion_type(self) -> Tuple[Type]:
        if self.default_conclusion is not None:
            return (type(self.default_conclusion),)
        return super().conclusion_type

    @staticmethod
    def start_rule_type() -> Type[Rule]:
        """
        :return: The type of the starting rule of the RDR classifier.
        """
        return SingleClassRule


class MultiClassRDR(RDRWithCodeWriter):
    """
    A multi class ripple down rules classifier, which can draw multiple conclusions for a case.
    This is done by going through all rules and checking if they fire or not, and adding stopping rules if needed,
    when wrong conclusions are made to stop these rules from firing again for similar cases.
    """
    evaluated_rules: Optional[List[Rule]] = None
    """
    The evaluated rules in the classifier for one case.
    """
    conclusions: Optional[List[CaseAttribute]] = None
    """
    The conclusions that the case belongs to.
    """
    stop_rule_conditions: Optional[CallableExpression] = None
    """
    The conditions of the stopping rule if needed.
    """
    mutually_exclusive: bool = False
    """
    The output of the classification of this rdr allows for more than one true value as conclusion.
    """

    def __init__(self, start_rule: Optional[MultiClassTopRule] = None,
                 mode: MCRDRMode = MCRDRMode.StopOnly, **kwargs):
        """
        :param start_rule: The starting rules for the classifier.
        :param mode: The mode of the classifier, either StopOnly or StopPlusRule, or StopPlusRuleCombined.
        """
        super(MultiClassRDR, self).__init__(start_rule, **kwargs)
        self.mode: MCRDRMode = mode

    def classify(self, case: Union[Case, SQLTable], modify_case: bool = False,
                 case_query: Optional[CaseQuery] = None) -> Set[Any]:
        evaluated_rule = self.start_rule
        self.conclusions = []
        while evaluated_rule:
            next_rule = evaluated_rule(case)
            if evaluated_rule.fired:
                rule_conclusion = evaluated_rule.conclusion(case)
                if evaluated_rule.corner_case_metadata is None and case_query is not None:
                    if rule_conclusion is not None and len(make_list(rule_conclusion)) > 0 \
                            and any(
                        ct in case_query.core_attribute_type for ct in map(type, make_list(rule_conclusion))):
                        evaluated_rule.corner_case_metadata = CaseFactoryMetaData.from_case_query(case_query)
                self.add_conclusion(rule_conclusion)
            evaluated_rule = next_rule
        return make_set(self.conclusions)

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None
                  , **kwargs) -> Set[Union[CaseAttribute, CallableExpression, None]]:
        """
        Classify a case, and ask the user for stopping rules or classifying rules if the classification is incorrect
         or missing by comparing the case with the target category if provided.

        :param case_query: The query containing the case to classify and the target category to compare the case with.
        :param expert: The expert to ask for differentiating features as new rule conditions or for extra conclusions.
        :return: The conclusions that the case belongs to.
        """
        self.conclusions = []
        self.stop_rule_conditions = None
        evaluated_rule = self.start_rule
        target_value = make_set(case_query.target_value)
        while evaluated_rule:
            next_rule = evaluated_rule(case_query.case)
            rule_conclusion = evaluated_rule.conclusion(case_query.case)

            if evaluated_rule.fired:
                if not make_set(rule_conclusion).issubset(target_value):
                    # Rule fired and conclusion is different from target
                    self.stop_wrong_conclusion_else_add_it(case_query, expert, evaluated_rule)
                else:
                    # Rule fired and target is correct or there is no target to compare
                    self.add_conclusion(rule_conclusion)

            if not next_rule:
                if not make_set(target_value).issubset(make_set(self.conclusions)):
                    # Nothing fired and there is a target that should have been in the conclusions
                    self.add_rule_for_case(case_query, expert)
                    # Have to check all rules again to make sure only this new rule fires
                    next_rule = self.start_rule
            evaluated_rule = next_rule
        return self.conclusions

    def write_rules_as_source_code_to_file(self, rule: Union[MultiClassTopRule, MultiClassStopRule],
                                           filename: str, parent_indent: str = "", defs_file: Optional[str] = None,
                                           cases_file: Optional[str] = None, package_name: Optional[str] = None):
        if rule == self.start_rule:
            with open(filename, "a") as file:
                file.write(f"{parent_indent}conclusions = set()\n")
        if rule.conditions:
            rule.write_corner_case_as_source_code(cases_file, package_name=package_name)
            if_clause = rule.write_condition_as_source_code(parent_indent, defs_file)
            with open(filename, "a") as file:
                file.write(if_clause)
            conclusion_indent = parent_indent
            if hasattr(rule, "refinement") and rule.refinement:
                self.write_rules_as_source_code_to_file(rule.refinement, filename, parent_indent + "    ",
                                                        defs_file=defs_file, cases_file=cases_file,
                                                        package_name=package_name)
                conclusion_indent = parent_indent + " " * 4
                with open(filename, "a") as file:
                    file.write(f"{conclusion_indent}else:\n")

            conclusion_call = rule.write_conclusion_as_source_code(conclusion_indent, defs_file)
            with open(filename, "a") as file:
                file.write(conclusion_call)

            if rule.alternative:
                self.write_rules_as_source_code_to_file(rule.alternative, filename, parent_indent, defs_file=defs_file,
                                                        cases_file=cases_file, package_name=package_name)

    @property
    def conclusion_type_hint(self) -> str:
        conclusion_types = [ct.__name__ for ct in self.conclusion_type if ct not in [list, set]]
        if len(conclusion_types) == 1:
            return f"Set[{conclusion_types[0]}]"
        else:
            return f"Set[Union[{', '.join(conclusion_types)}]]"

    def _get_types_to_import(self) -> Tuple[Set[Type], Set[Type], Set[Type]]:
        main_types, defs_types, cases_types = super()._get_types_to_import()
        main_types.update({Set, make_set})
        defs_types.update({List, Set})
        return main_types, defs_types, cases_types

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if not self.start_rule:
            conditions = expert.ask_for_conditions(case_query)
            self.start_rule = MultiClassTopRule.from_case_query(case_query)

    @property
    def last_top_rule(self) -> Optional[MultiClassTopRule]:
        """
        Get the last top rule in the tree.
        """
        if not self.start_rule.furthest_alternative:
            return self.start_rule
        else:
            return self.start_rule.furthest_alternative[-1]

    def stop_wrong_conclusion_else_add_it(self, case_query: CaseQuery, expert: Expert,
                                          evaluated_rule: MultiClassTopRule):
        """
        Stop a wrong conclusion by adding a stopping rule.
        """
        rule_conclusion = evaluated_rule.conclusion(case_query.case)
        if is_conflicting(rule_conclusion, case_query.target_value):
            self.stop_conclusion(case_query, expert, evaluated_rule)
        else:
            self.add_conclusion(rule_conclusion)

    def stop_conclusion(self, case_query: CaseQuery,
                        expert: Expert, evaluated_rule: MultiClassTopRule):
        """
        Stop a conclusion by adding a stopping rule.

        :param case_query: The case query to stop the conclusion for.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :param evaluated_rule: The evaluated rule to ask the expert about.
        """
        conditions = expert.ask_for_conditions(case_query, evaluated_rule)
        evaluated_rule.fit_rule(case_query)
        if self.mode == MCRDRMode.StopPlusRule:
            self.stop_rule_conditions = conditions
        if self.mode == MCRDRMode.StopPlusRuleCombined:
            new_top_rule_conditions = conditions.combine_with(evaluated_rule.conditions)
            case_query.conditions = new_top_rule_conditions
            self.add_top_rule(case_query)

    def add_rule_for_case(self, case_query: CaseQuery, expert: Expert):
        """
        Add a rule for a case that has not been classified with any conclusion.

        :param case_query: The case query to add the rule for.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if self.stop_rule_conditions and self.mode == MCRDRMode.StopPlusRule:
            conditions = self.stop_rule_conditions
            self.stop_rule_conditions = None
            case_query.conditions = conditions
        else:
            conditions = expert.ask_for_conditions(case_query)
        self.add_top_rule(case_query)

    def add_conclusion(self, rule_conclusion: List[Any]) -> None:
        """
        Add the conclusion of the evaluated rule to the list of conclusions.

        :param rule_conclusion: The conclusion of the evaluated rule, which can be a single conclusion
         or a set of conclusions.
        """
        conclusion_types = [type(c) for c in self.conclusions]
        if type(rule_conclusion) not in conclusion_types:
            self.conclusions.extend(make_list(rule_conclusion))
        else:
            same_type_conclusions = [c for c in self.conclusions if type(c) == type(rule_conclusion)]
            combined_conclusion = rule_conclusion if isinstance(rule_conclusion, set) \
                else {rule_conclusion}
            combined_conclusion = copy(combined_conclusion)
            for c in same_type_conclusions:
                combined_conclusion.update(c if isinstance(c, set) else make_set(c))
                self.conclusions.remove(c)
            self.conclusions.extend(make_list(combined_conclusion))

    def add_top_rule(self, case_query: CaseQuery):
        """
        Add a top rule to the classifier, which is a rule that is always checked and is part of the start_rules list.

        :param case_query: The case query to add the top rule for.
        """
        self.start_rule.alternative = MultiClassTopRule.from_case_query(case_query)

    @staticmethod
    def start_rule_type() -> Type[Rule]:
        """
        :return: The type of the starting rule of the RDR classifier.
        """
        return MultiClassTopRule


class GeneralRDR(RippleDownRules):
    """
    A general ripple down rules classifier, which can draw multiple conclusions for a case, but each conclusion is part
    of a set of mutually exclusive conclusions. Whenever a conclusion is made, the classification restarts from the
    starting rule, and all the rules that belong to the class of the made conclusion are not checked again. This
    continues until no more rules can be fired. In addition, previous conclusions can be used as conditions or input to
    the next classification/cycle.
    Another possible mode is to have rules that are considered final, when fired, inference will not be restarted,
     and only a refinement can be made to the final rule, those can also be used in another SCRDR of their own that
     gets called when the final rule fires.
    """

    def __init__(self, category_rdr_map: Optional[Dict[str, Union[SingleClassRDR, MultiClassRDR]]] = None,
                 **kwargs):
        """
        :param category_rdr_map: A map of case attribute names to ripple down rules classifiers,
        where each category is a parent category that has a set of mutually exclusive (in case of SCRDR) child
        categories, e.g. {'species': SCRDR, 'habitats': MCRDR}, where 'species' and 'habitats' are attribute names
        for a case of type Animal, while SCRDR and MCRDR are SingleClass and MultiClass ripple down rules classifiers.
        Species can have values like Mammal, Bird, Fish, etc. which are mutually exclusive, while Habitat can have
        values like Land, Water, Air, etc., which are not mutually exclusive due to some animals living more than one
        habitat.
        """
        self.start_rules_dict: Dict[str, Union[SingleClassRDR, MultiClassRDR]] \
            = category_rdr_map if category_rdr_map else {}
        super(GeneralRDR, self).__init__(**kwargs)
        self.all_figs: List[Figure] = [sr.fig for sr in self.start_rules_dict.values()]

    def add_rdr(self, rdr: Union[SingleClassRDR, MultiClassRDR], case_query: Optional[CaseQuery] = None):
        """
        Add a ripple down rules classifier to the map of classifiers.

        :param rdr: The ripple down rules classifier to add.
        :param case_query: The case query to add the classifier for.
        """
        name = case_query.attribute_name if case_query else rdr.name
        self.start_rules_dict[name] = rdr

    @property
    def start_rule(self) -> Optional[Union[SingleClassRule, MultiClassTopRule]]:
        return self.start_rules[0] if self.start_rules_dict else None

    @start_rule.setter
    def start_rule(self, value: Union[SingleClassRDR, MultiClassRDR]):
        if value:
            self.start_rules_dict[value.attribute_name] = value

    @property
    def start_rules(self) -> List[Union[SingleClassRule, MultiClassTopRule]]:
        return [rdr.start_rule for rdr in self.start_rules_dict.values()]

    def classify(self, case: Any, modify_case: bool = False,
                 case_query: Optional[CaseQuery] = None) -> Optional[Dict[str, Any]]:
        """
        Classify a case by going through all RDRs and adding the categories that are classified, and then restarting
        the classification until no more categories can be added.

        :param case: The case to classify.
        :param modify_case: Whether to modify the original case or create a copy and modify it.
        :param case_query: The case query containing the case and the target category to compare the case with.
        :return: The categories that the case belongs to.
        """
        return general_rdr_classify(self.start_rules_dict, case, modify_original_case=modify_case,
                                    case_query=case_query)

    def _fit_case(self, case_query: CaseQuery, expert: Optional[Expert] = None, **kwargs) \
            -> Dict[str, Any]:
        """
        Fit the GRDR on a case, if the target is a new type of category, a new RDR is created for it,
        else the existing RDR of that type will be fitted on the case, and then classification is done and all
        concluded categories are returned. If the category is mutually exclusive, an SCRDR is created, else an MCRDR.
        In case of SCRDR, multiple conclusions of the same type replace each other, in case of MCRDR, they are added if
        they are accepted by the expert, and the attribute of that category is represented in the case as a set of
        values.

        :param case_query: The query containing the case to classify and the target category to compare the case
        with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        :return: The categories that the case belongs to.
        """
        case_query_cp = copy(case_query)
        self.classify(case_query_cp.case, modify_case=True)
        case_query_cp.update_target_value()

        self.start_rules_dict[case_query_cp.attribute_name].fit_case(case_query_cp, expert, **kwargs)

        return self.classify(case_query.case)

    def update_start_rule(self, case_query: CaseQuery, expert: Expert):
        """
        Update the starting rule of the classifier.

        :param case_query: The case query to update the starting rule with.
        :param expert: The expert to ask for differentiating features as new rule conditions.
        """
        if case_query.attribute_name not in self.start_rules_dict:
            new_rdr = self.initialize_new_rdr_for_attribute(case_query)
            self.add_rdr(new_rdr, case_query)

    @staticmethod
    def initialize_new_rdr_for_attribute(case_query: CaseQuery):
        """
        Initialize the appropriate RDR type for the target.
        """
        return SingleClassRDR(default_conclusion=case_query.default_value) if case_query.mutually_exclusive \
            else MultiClassRDR()

    def _to_json(self) -> Dict[str, Any]:
        return {"start_rules": {name: rdr.to_json() for name, rdr in self.start_rules_dict.items()}
            , "generated_python_file_name": self.generated_python_file_name,
                "name": self.name,
                "case_type": get_full_class_name(self.case_type) if self.case_type is not None else None,
                "case_name": self.case_name}

    @classmethod
    def _from_json(cls, data: Dict[str, Any]) -> GeneralRDR:
        """
        Create an instance of the class from a json
        """
        start_rules_dict = {}
        for k, v in data["start_rules"].items():
            start_rules_dict[k] = get_type_from_string(v['_type']).from_json(v)
        new_rdr = cls(category_rdr_map=start_rules_dict)
        if "generated_python_file_name" in data:
            new_rdr.generated_python_file_name = data["generated_python_file_name"]
        if "name" in data:
            new_rdr.name = data["name"]
        if "case_type" in data:
            new_rdr.case_type = get_type_from_string(data["case_type"])
        if "case_name" in data:
            new_rdr.case_name = data["case_name"]
        return new_rdr

    def update_from_python(self, model_dir: str, package_name: Optional[str] = None) -> None:
        """
        Update the rules from the generated python file, that might have been modified by the user.

        :param model_dir: The directory where the model is stored.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        """
        for rdr in self.start_rules_dict.values():
            rdr.update_from_python(model_dir, package_name=package_name)

    def _write_to_python(self, model_dir: str, package_name: Optional[str] = None) -> None:
        """
        Write the tree of rules as source code to a file.

        :param model_dir: The directory where the model is stored.
        :param relative_imports: Whether to use relative imports in the generated python file.
        """
        for rdr in self.start_rules_dict.values():
            rdr._write_to_python(model_dir, package_name=package_name)
        func_def = f"def classify(case: {self.case_type.__name__}, **kwargs) -> {self.conclusion_type_hint}:\n"
        file_path = model_dir + f"/{self.generated_python_file_name}.py"
        with open(file_path, "w") as f:
            f.write(self._get_imports(file_path=file_path, package_name=package_name) + "\n\n")
            f.write("classifiers_dict = dict()\n")
            for rdr_key, rdr in self.start_rules_dict.items():
                f.write(f"classifiers_dict['{rdr_key}'] = {self.rdr_key_to_function_name(rdr_key)}\n")
            f.write("\n\n")
            f.write(func_def)
            f.write(f"{' ' * 4}if not isinstance(case, Case):\n"
                    f"{' ' * 4}    case = create_case(case, max_recursion_idx=3)\n""")
            f.write(f"{' ' * 4}return general_rdr_classify(classifiers_dict, case, **kwargs)\n")

    @property
    def _default_generated_python_file_name(self) -> Optional[str]:
        """
        :return: The default generated python file name.
        """
        if self.start_rule is None or self.start_rule.conclusion is None:
            return None
        return f"{str_to_snake_case(self.case_name)}_rdr".lower()

    @property
    def conclusion_type_hint(self) -> str:
        return f"Dict[str, Any]"

    def _get_imports(self, file_path: Optional[str] = None, package_name: Optional[str] = None) -> str:
        """
        Get the imports needed for the generated python file.

        :param file_path: The path to the file where the imports will be written, if None, the imports will be absolute.
        :param package_name: The name of the package that contains the RDR classifier function, this
        is required in case of relative imports in the generated python file.
        :return: The imports needed for the generated python file.
        """
        all_types = set()
        # add type hints
        all_types.update({Dict, Any})
        # import rdr type
        all_types.add(general_rdr_classify)
        # add case type
        all_types.update({Case, create_case, self.case_type})
        # get the imports from the types
        imports = get_imports_from_types(all_types, target_file_path=file_path, package_name=package_name)
        # add rdr python generated functions.
        for rdr_key, rdr in self.start_rules_dict.items():
            imports.append(
                f"from . import {rdr.generated_python_file_name} as {self.rdr_key_to_function_name(rdr_key)}")
        return '\n'.join(imports)

    @staticmethod
    def rdr_key_to_function_name(rdr_key: str) -> str:
        """
        Convert the RDR key to a function name.

        :param rdr_key: The RDR key to convert.
        :return: The function name.
        """
        return rdr_key.replace(".", "_").lower() + "_classifier"



from __future__ import annotations

import inspect
import typing
from dataclasses import dataclass, field

import typing_extensions
from omegaconf import MISSING
from sqlalchemy.orm import DeclarativeBase as SQLTable
from typing_extensions import Any, Optional, Dict, Type, Tuple, Union, List, get_origin, Set, Callable

from ..utils import get_method_name, get_function_import_data, get_function_representation
from .callable_expression import CallableExpression
from .case import create_case, Case
from ..utils import copy_case, make_list, make_set, get_origin_and_args_from_type_hint, get_value_type_from_type_hint, \
    typing_to_python_type


@dataclass
class CaseQuery:
    """
    This is a dataclass that represents an attribute of an object and its target value. If attribute name is
    not provided, it will be inferred from the attribute itself or from the attribute type or from the target value,
    depending on what is provided.
    """
    original_case: Any
    """
    The case that the attribute belongs to.
    """
    attribute_name: str
    """
    The name of the attribute.
    """
    _attribute_types: Tuple[Type]
    """
    The type(s) of the attribute.
    """
    mutually_exclusive: bool
    """
    Whether the attribute can only take one value (i.e. True) or multiple values (i.e. False).
    """
    case_factory: Optional[Callable[[], Any]] = None
    """
    The factory method that can be used to recreate the original case.
    """
    case_factory_idx: Optional[int] = None
    """
    This is used when the case factory is a list of cases, this index is used to select the case from the list.
    """
    case_conf: Optional[CaseConf] = None
    """
    The case configuration that is used to (re)create the original case, recommended to be used when you want to 
    the case to persist in the rule base, this would allow it to be used for merging with other similar conclusion RDRs.
    """
    scenario: Optional[Callable] = None
    """
    The executable scenario is the root callable that recreates the situation that the case is 
    created in, for example, when the case is created from a test function, this would be the test function itself.
    """
    _target: Optional[CallableExpression] = None
    """
    The target expression of the attribute.
    """
    default_value: Optional[Any] = None
    """
    The default value of the attribute. This is used when the target value is not provided.
    """
    scope: Optional[Dict[str, Any]] = field(default_factory=lambda: inspect.currentframe().f_back.f_back.f_globals)
    """
    The global scope of the case query. This is used to evaluate the conditions and prediction, and is what is available
    to the user when they are prompted for input. If it is not provided, it will be set to the global scope of the
    caller.
    """
    _case: Optional[Union[Case, SQLTable]] = None
    """
    The created case from the original case that the attribute belongs to.
    """
    _target_value: Optional[Any] = None
    """
    The target value of the case query. (This is the result of the target expression evaluation on the case.)
    """
    conditions: Optional[CallableExpression] = None
    """
    The conditions that must be satisfied for the target value to be valid.
    """
    is_function: bool = False
    """
    Whether the case is a dict representing the arguments of an actual function or not,
    most likely means it came from RDRDecorator, the the rdr takes function arguments and outputs the function output.
    """
    function_args_type_hints: Optional[Dict[str, Type]] = None
    """
    The type hints of the function arguments. This is used to recreate the function signature.
    """

    @property
    def case_type(self) -> Type:
        """
        :return: The type of the case that the attribute belongs to.
        """
        if self.is_function:
            if self.function_args_type_hints is not None:
                func_args = [arg for name, arg in self.function_args_type_hints.items() if name != 'return']
                case_type_args = Union[tuple(func_args)]
            else:
                case_type_args = Any
            return Dict[str, case_type_args]
        else:
            return self.original_case._obj_type if isinstance(self.original_case, Case) else type(self.original_case)

    @property
    def case(self) -> Any:
        """
        :return: The case that the attribute belongs to.
        """
        if self._case is not None:
            return self._case
        elif not isinstance(self.original_case, Case):
            self._case = create_case(self.original_case, max_recursion_idx=3)
        else:
            self._case = self.original_case
        return self._case

    @case.setter
    def case(self, value: Any):
        """
        Set the case that the attribute belongs to.
        """
        if not isinstance(value, (Case, SQLTable)):
            raise ValueError("The case must be a Case or SQLTable object.")
        self._case = value

    @property
    def attribute_type_hint(self) -> str:
        """
        :return: The type hint of the attribute as a typing object.
        """
        if len(self.core_attribute_type) > 1:
            attribute_types_str = f"Union[{', '.join([t.__name__ for t in self.core_attribute_type])}]"
        else:
            attribute_types_str = self.core_attribute_type[0].__name__
        if all(t in self.attribute_type for t in [list, set]) and len(self.core_attribute_type) > 2:
            return f"List[{attribute_types_str}]"
        else:
            return attribute_types_str

    @property
    def core_attribute_type(self) -> Tuple[Type]:
        """
        :return: The core type of the attribute.
        """
        return tuple(t for t in self.attribute_type if t not in (set, list))

    @property
    def attribute_type(self) -> Tuple[Type]:
        """
        :return: The type of the attribute.
        """
        if not isinstance(self._attribute_types, tuple):
            self._attribute_types = tuple(make_set(self._attribute_types))
        origin, args = get_origin_and_args_from_type_hint(self._attribute_types)
        if origin is not None:
            att_types = make_set(origin)
            if origin in (list, set, tuple, List, Set, Union, Tuple):
                att_types.update(make_set(args))
            elif origin in (dict, Dict):
                # ignore the key type
                if args and len(args) > 1:
                    att_types.update(make_set(args[1]))
            self._attribute_types = tuple(att_types)
        if not self.mutually_exclusive and (list not in self._attribute_types):
            self._attribute_types = tuple(make_list(self._attribute_types) + [set, list])
        return self._attribute_types

    @attribute_type.setter
    def attribute_type(self, value: Type):
        """
        Set the type of the attribute.
        """
        self._attribute_types = tuple(make_list(value))

    @property
    def name(self):
        """
        :return: The name of the case query.
        """
        return f"{self.case_name}.{self.attribute_name}"

    @property
    def case_name(self) -> str:
        """
        :return: The name of the case.
        """
        return self.case._name if isinstance(self.case, Case) else self.case.__class__.__name__

    @property
    def target(self) -> Optional[CallableExpression]:
        """
        :return: The target expression of the attribute.
        """
        if (self._target is not None) and (not isinstance(self._target, CallableExpression)):
            self._target = CallableExpression(conclusion=self._target, conclusion_type=self.attribute_type,
                                              scope=self.scope, mutually_exclusive=self.mutually_exclusive)
        return self._target

    @target.setter
    def target(self, value: Optional[CallableExpression]):
        """
        Set the target expression of the attribute.
        """
        if value is not None and not isinstance(value, (CallableExpression, str)):
            raise ValueError("The target must be a CallableExpression or a string.")
        self._target = value
        self.update_target_value()

    @property
    def target_value(self) -> Any:
        """
        :return: The target value of the case query.
        """
        if self._target_value is None:
            self.update_target_value()
        return self._target_value

    def update_target_value(self):
        """
        Update the target value of the case query.
        """
        if isinstance(self.target, CallableExpression):
            self._target_value = self.target(self.case)
        else:
            self._target_value = self.target

    def __str__(self):
        header = f"CaseQuery: {self.name}"
        target = f"Target: {self.name} |= {self.target if self.target is not None else '?'}"
        conditions = f"Conditions: {self.conditions if self.conditions is not None else '?'}"
        return "\n".join([header, target, conditions])

    def __repr__(self):
        return self.__str__()

    def __copy__(self):
        return CaseQuery(self.original_case, self.attribute_name, self.attribute_type,
                         self.mutually_exclusive, _target=self.target, default_value=self.default_value,
                         scope=self.scope, _case=copy_case(self.case), _target_value=self.target_value,
                         conditions=self.conditions, is_function=self.is_function,
                         function_args_type_hints=self.function_args_type_hints,
                         case_factory=self.case_factory, case_factory_idx=self.case_factory_idx,
                         case_conf=self.case_conf, scenario=self.scenario)


@dataclass
class CaseConf:
    factory_method: Callable[[Any], Any] = MISSING

    def create(self) -> Any:
        return self.factory_method()


@dataclass
class CaseFactoryMetaData:
    factory_method: Optional[Callable[[Optional[CaseConf]], Any]] = None
    factory_idx: Optional[int] = None
    case_conf: Optional[CaseConf] = None
    scenario: Optional[Callable] = None

    @classmethod
    def from_case_query(cls, case_query: CaseQuery) -> CaseFactoryMetaData:
        return cls(factory_method=case_query.case_factory, factory_idx=case_query.case_factory_idx,
                   case_conf=case_query.case_conf, scenario=case_query.scenario)

    def __repr__(self):
        factory_method_repr = None
        scenario_repr = None
        if self.factory_method is not None:
            factory_method_repr = get_function_representation(self.factory_method)
        if self.scenario is not None:
            scenario_repr = get_function_representation(self.scenario)
        return (f"CaseFactoryMetaData("
                f"factory_method={factory_method_repr}, "
                f"factory_idx={self.factory_idx}, "
                f"case_conf={self.case_conf},"
                f" scenario={scenario_repr})")

    def __str__(self):
        return self.__repr__()


import os
import unittest

from .datasets import load_zoo_dataset, Species
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.user_interface.object_diagram import generate_object_graph


class Address:
    def __init__(self, city):
        self.city = city


class Person:
    def __init__(self, name, address):
        self.name = name
        self.address = address


class ObjectDiagramTestCase(unittest.TestCase):
    """Test case for generating object diagrams of the ripple down rules package."""
    cases: list
    targets: list
    # cq: CaseQuery
    person: Person

    @classmethod
    def setUpClass(cls):
        cls.cases, cls.targets = load_zoo_dataset(cache_file=f"{os.path.dirname(__file__)}/test_results/zoo")
        cls.cq = CaseQuery(cls.cases[0], "species", (Species,), True, _target=cls.targets[0])
        cls.person = Person("Ahmed", Address("Cairo"))

    def test_generate_person_diagram(self):
        # Generate object diagram for the Person instance
        graph = generate_object_graph(self.person)
        graph.render(f'{os.path.dirname(__file__)}/test_results/person_object_diagram', view=False)
        self.assertTrue(os.path.isfile(f'{os.path.dirname(__file__)}/test_results/person_object_diagram.svg'))

    def test_generate_case_query_diagram(self):
        # Generate object diagram for the CaseQuery instance
        graph = generate_object_graph(self.cq)
        graph.render(f'{os.path.dirname(__file__)}/test_results/case_query_object_diagram', view=False)
        self.assertTrue(os.path.isfile(f'{os.path.dirname(__file__)}/test_results/case_query_object_diagram.svg'))



import importlib
import sys
from os.path import dirname

import pytest
from typing_extensions import Callable

try:
    from PyQt6.QtWidgets import QApplication
    from ripple_down_rules.user_interface.gui import RDRCaseViewer
except ImportError as e:
    QApplication = None
    RDRCaseViewer = None

from ..conf.world.handles_and_containers import HandlesAndContainersWorld
from ..datasets import *
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.experts import Human
from ripple_down_rules.helpers import is_matching
from ripple_down_rules.rdr import GeneralRDR

app: Optional[QApplication] = None
viewer: Optional[RDRCaseViewer] = None

if RDRCaseViewer is not None and QApplication is not None:
    app = QApplication(sys.argv)
    viewer = RDRCaseViewer(save_dir="./test_generated_rdrs")


def handles_and_containers_world() -> World:
    return HandlesAndContainersWorld().create()

def get_possible_drawers() -> List[Drawer]:
    """
    Get all possible drawer types.
    """
    all_possible_drawers = []
    world = handles_and_containers_world()
    for handle in [body for body in world.bodies if isinstance(body, Handle)]:
        for container in [body for body in world.bodies if isinstance(body, Container)]:
            view = Drawer(handle, container, world=world)
            all_possible_drawers.append(view)
    return all_possible_drawers

@pytest.fixture
def drawer_case_queries() -> List[CaseQuery]:
    case_queries = [CaseQuery(possible_drawer, "correct", (bool,), True, default_value=False,
                              case_factory=get_possible_drawers, case_factory_idx=i)
                    for i, possible_drawer in enumerate(get_possible_drawers())]
    return case_queries


@dataclass
class ExpertConfig:
    filename: str
    use_loaded_answers: bool


@pytest.fixture
def expert():
    def _create(conf: ExpertConfig):
        """
        Fixture to create an expert.

        :param conf: ExpertConfig object containing configuration for the expert.
        """
        human = Human(use_loaded_answers=conf.use_loaded_answers, viewer=viewer)
        filename = os.path.join(dirname(__file__), "../test_expert_answers", conf.filename)
        human.load_answers(filename)
        return human

    return _create


@pytest.fixture
def drawer_cabinet_expert(expert) -> Human:
    """
    Fixture to create an expert for drawer and cabinet views.
    """
    conf = ExpertConfig("drawer_cabinet_expert_answers_fit", use_loaded_answers=True)
    return expert(conf)


@pytest.fixture
def drawer_expert(expert) -> Human:
    """
    Fixture to create an expert for drawer views.
    """
    conf = ExpertConfig("correct_drawer_rdr_expert_answers_fit", use_loaded_answers=True)
    return expert(conf)


@pytest.fixture
def drawer_case_query() -> CaseQuery:
    """
    Create a CaseQuery for the given view type in the provided world.
    """
    return CaseQuery(handles_and_containers_world(), "views", (Drawer,), False,
                     case_factory=handles_and_containers_world)


def possibilities_rdr_verification(rdr: GeneralRDR, case_query: CaseQuery) -> None:
    conclusions = rdr.classify(case_query.original_case)[case_query.attribute_name]
    assert len([v for v in conclusions if isinstance(v, case_query.core_attribute_type)]) > 0


@pytest.fixture
def drawer_rdr(drawer_case_query, drawer_cabinet_expert) -> GeneralRDR:
    """
    Fixture to create a GeneralRDR for drawer views.
    """
    rdr = GeneralRDR()
    rdr.fit_case(drawer_case_query, expert=drawer_cabinet_expert)
    possibilities_rdr_verification(rdr, drawer_case_query)
    return rdr

def get_method_object_from_pytest_request(request) -> Callable:
    test_module = request.module.__name__  # e.g., "test_my_module"
    test_class = request.cls.__name__ if request.cls else None  # if inside a class
    test_name = request.node.name
    func = importlib.import_module(test_module)
    if test_class:
        func = getattr(getattr(func, test_class), test_name)
    else:
        func = getattr(func, test_name)
    return func

@pytest.fixture
def drawer_cabinet_rdr(request, drawer_cabinet_expert) -> GeneralRDR:
    world = handles_and_containers_world()
    rdr = GeneralRDR()
    for view in [Drawer, Cabinet]:
        rdr.fit_case(CaseQuery(world, "views", (view,), False, case_factory=handles_and_containers_world),
                     expert=drawer_cabinet_expert, scenario=get_method_object_from_pytest_request(request))
    found_views = rdr.classify(world)
    for view in [Drawer, Cabinet]:
        assert len([v for v in found_views["views"] if isinstance(v, view)]) > 0
    return rdr


@pytest.fixture
def correct_drawer_rdr(drawer_case_queries, drawer_expert) -> GeneralRDR:
    rdr = GeneralRDR()
    rdr.fit(drawer_case_queries, expert=drawer_expert, animate_tree=False)
    for case_query in drawer_case_queries:
        assert is_matching(rdr.classify, case_query)
    return rdr



from typing_extensions import Dict, List, Set, Union
from ...datasets import PhysicalObject, Robot


def conditions_164855806603893754507167918997373216146(case) -> bool:
    def conditions_for_physical_object_select_objects_that_are_parts_of_robot(self_: PhysicalObject, objects: List[PhysicalObject], robot: Robot, output_: PhysicalObject) -> bool:
        """Get conditions on whether it's possible to conclude a value for PhysicalObject_select_objects_that_are_parts_of_robot.output_  of type PhysicalObject."""
        return robot is not None
    return conditions_for_physical_object_select_objects_that_are_parts_of_robot(**case)


def conclusion_164855806603893754507167918997373216146(case) -> List[PhysicalObject]:
    def physical_object_select_objects_that_are_parts_of_robot(self_: PhysicalObject, objects: List[PhysicalObject], robot: Robot, output_: PhysicalObject) -> List[PhysicalObject]:
        """Get possible value(s) for PhysicalObject_select_objects_that_are_parts_of_robot.output_  of type PhysicalObject."""
        robot_parts = [obj for obj in objects if obj in robot.parts]
        return robot_parts
    return physical_object_select_objects_that_are_parts_of_robot(**case)





from types import NoneType
from ...datasets import PhysicalObject, Robot
from typing_extensions import Dict, Optional, Union


def conditions_226969243620390858682731042391766665817(case) -> bool:
    def conditions_for_physical_object_is_a_robot(self_: PhysicalObject, output_: bool) -> bool:
        """Get conditions on whether it's possible to conclude a value for PhysicalObject_is_a_robot.output_  of type ."""
        return True
    return conditions_for_physical_object_is_a_robot(**case)


def conclusion_226969243620390858682731042391766665817(case) -> bool:
    def physical_object_is_a_robot(self_: PhysicalObject, output_: bool) -> bool:
        """Get possible value(s) for PhysicalObject_is_a_robot.output_  of type ."""
        return isinstance(self_, Robot)
    return physical_object_is_a_robot(**case)





import os
import sys

import pytest
from typing_extensions import Type

try:
    from PyQt6.QtWidgets import QApplication
    from ripple_down_rules.user_interface.gui import RDRCaseViewer
except ImportError as e:
    QApplication = None
    RDRCaseViewer = None

from .conf.world.handles_and_containers import HandlesAndContainersWorld
from .datasets import *
from ripple_down_rules.datastructures.dataclasses import CaseQuery
from ripple_down_rules.experts import Human
from ripple_down_rules.helpers import is_matching
from ripple_down_rules.rdr import GeneralRDR


app: Optional[QApplication] = None
viewer: Optional[RDRCaseViewer] = None

if RDRCaseViewer is not None and QApplication is not None:
    app = QApplication(sys.argv)
    viewer = RDRCaseViewer(save_dir="./test_generated_rdrs")


def pytest_generate_tests(metafunc):

    if metafunc.definition.originalname == "test_should_i_ask_the_expert_for_a_target":
        all_cases, all_targets = load_zoo_dataset("./test_results/zoo")

        possible_case_queries = [CaseQuery(all_cases[0], 'species', (Species,), True),
                                 CaseQuery(all_cases[1], 'habitat', (Habitat,), False)]
        metafunc.parametrize("case_query", possible_case_queries)

        possible_conclusions = [[], None, True, False, set(), {}, {'species': Species.mammal}, Species.mammal,
                                Habitat.land, {'habitat': Habitat.water}, [Habitat.water, Habitat.land],
                                {'species': Species.fish, 'habitat': Habitat.water}]
        metafunc.parametrize("conclusions", possible_conclusions)

        possible_update_existing = [True, False]
        metafunc.parametrize("update_existing", possible_update_existing)


@pytest.fixture
def handles_and_containers_world() -> World:
    return HandlesAndContainersWorld().create()


@pytest.fixture
def drawer_case_queries(handles_and_containers_world) -> List[CaseQuery]:
    all_possible_drawers = []
    world = handles_and_containers_world
    for handle in [body for body in world.bodies if isinstance(body, Handle)]:
        for container in [body for body in world.bodies if isinstance(body, Container)]:
            view = Drawer(handle, container, world=world)
            all_possible_drawers.append(view)
    case_queries = [CaseQuery(possible_drawer, "correct", (bool,), True, default_value=False)
                               for possible_drawer in all_possible_drawers]
    return case_queries


@pytest.fixture
def view_rdr(handles_and_containers_world, views=(Drawer, Cabinet),
             use_loaded_answers: bool = True,
             save_answers: bool = False,
             append: bool = False) -> GeneralRDR:
    world = handles_and_containers_world
    expert = Human(use_loaded_answers=use_loaded_answers, append=append, viewer=viewer)
    filename = os.path.join(os.getcwd(), "test_expert_answers/view_rdr_expert_answers_fit")
    if use_loaded_answers:
        expert.load_answers(filename)
    rdr = GeneralRDR()
    for view in views:
        rdr.fit_case(CaseQuery(world, "views", (view,), False), expert=expert)
    if save_answers:
        expert.save_answers(filename)

    found_views = rdr.classify(world)
    for view in views:
        assert len([v for v in found_views["views"] if isinstance(v, view)]) > 0

    return rdr


@pytest.fixture
def view_case_query(world: World, view_type: Type[View]) -> CaseQuery:
    """
    Create a CaseQuery for the given view type in the provided world.
    """
    return CaseQuery(world, "views", (view_type,), False)


@pytest.fixture
def drawer_rdr(drawer_case_queries, use_loaded_answers: bool = True, save_answers: bool = False):
    expert = Human(use_loaded_answers=use_loaded_answers)
    filename = os.path.join(os.getcwd(), "test_expert_answers/correct_drawer_rdr_expert_answers_fit")
    if use_loaded_answers:
        expert.load_answers(filename)
    rdr = GeneralRDR()
    rdr.fit(drawer_case_queries, expert=expert, animate_tree=False)
    if save_answers:
        expert.save_answers(filename)
    for case_query in drawer_case_queries:
        assert is_matching(rdr.classify, case_query)
    return rdr



# ===================== Possible World Configurations ========================
from dataclasses import dataclass, field

from typing_extensions import List, Callable

from .base_config import WorldConf, BodyConf, Connection, FixedConnectionConf, PrismaticConnectionConf, \
    ContainerConf, HandleConf

from ...factories.world.handles_and_containers import create_world_with_handles_and_containers


@dataclass
class Handle1(HandleConf):
    name: str = "Handle1"


@dataclass
class Handle2(HandleConf):
    name: str = "Handle2"


@dataclass
class Container1(ContainerConf):
    name: str = "Container1"


@dataclass
class Container2(ContainerConf):
    name: str = "Container2"


def bodies():
    return [
        Handle1(),
        Handle2(),
        Container1(),
        Container2()
    ]


@dataclass
class HandlesAndContainersWorld(WorldConf):
    bodies: List[BodyConf] = field(default_factory=bodies, init=False)
    connections: List[Connection] = field(default_factory=lambda: [
        FixedConnectionConf(parent=Container1(), child=Handle1()),
        PrismaticConnectionConf(parent=Container2(), child=Container1())], init=False)
    factory_method: Callable = field(default=create_world_with_handles_and_containers, init=False)


